#include <assert.h>
#include <stdlib.h>
#include "set.h"

typedef struct Entry {
   stralloc s;
   struct Entry* next;
} Entry;

struct Set {
   unsigned int size;
   Entry** bucket;
   /* iterator */
   unsigned int next_index;
   Entry* next_entry;
};
typedef struct Set Set;

Set* create_set(unsigned int size) {
   assert(size > 0);
   Set* set = calloc(1, sizeof(Set));
   if (!set) return 0;
   set->size = size;
   set->bucket = calloc(size, sizeof(struct Entry*));
   if (!set->bucket) {
      free(set);
      return 0;
   }
   return set;
}

/* hash algorithm stolen from Dan Bernstein's cdbhash.c */
#define HASHSTART 5381

static unsigned int compute_hash(stralloc* s) {
   int hashval = HASHSTART;
   for (unsigned int index = 0; index < s->len; ++index) {
      char ch = s->s[index];
      hashval += hashval << 5;
      hashval ^= ch;
   }
   return hashval;
}

bool set_add(Set* set, stralloc* s) {
   unsigned int index = compute_hash(s) % set->size;
   Entry* entry = set->bucket[index];
   while (entry && stralloc_diff(s, &entry->s) != 0) {
      entry = entry->next;
   }
   if (entry) return true; /* already in set */
   entry = calloc(1, sizeof(Entry));
   if (!entry) return false;
   if (!stralloc_copy(&entry->s, s)) {
      free(entry); return false;
   }
   entry->next = set->bucket[index];
   set->bucket[index] = entry;
   return true;
}

bool in_set(Set* set, stralloc* s) {
   unsigned int index = compute_hash(s) % set->size;
   Entry* entry = set->bucket[index];
   while (entry && stralloc_diff(s, &entry->s) != 0) {
      entry = entry->next;
   }
   return entry;
}

void set_first(Set* set) {
   set->next_index = 0; set->next_entry = 0;
}

const stralloc* set_next(Set* set) {
   Entry* entry = set->next_entry;
   while (!entry && set->next_index < set->size) {
      entry = set->bucket[set->next_index++];
   }
   if (!entry) return 0;
   set->next_entry = entry->next;
   return &entry->s;
}
